@@ -1,10 +1,11 @@
-from __future__ import absolute_import, division, print_function
+from __future__ import annotations
 
 import io
 import logging
 import os
 import re
 from glob import has_magic
+from pathlib import Path
 
 # for backwards compat, we export cache things from here too
 from .caching import (  # noqa: F401
@@ -22,7 +23,6 @@ from .utils import (
     build_name_function,
     infer_compression,
     stringify_path,
-    update_storage_options,
 )
 
 logger = logging.getLogger("fsspec")
@@ -95,7 +95,7 @@ class OpenFile:
         )
 
     def __repr__(self):
-        return "<OpenFile '{}'>".format(self.path)
+        return f"<OpenFile '{self.path}'>"
 
     def __enter__(self):
         mode = self.mode.replace("t", "").replace("b", "") + "b"
@@ -198,7 +198,7 @@ class OpenFiles(list):
         return out
 
     def __repr__(self):
-        return "<List of %s OpenFile instances>" % len(self)
+        return f"<List of {len(self)} OpenFile instances>"
 
 
 def open_files(
@@ -269,6 +269,16 @@ def open_files(
     -------
     An ``OpenFiles`` instance, which is a list of ``OpenFile`` objects that can
     be used as a single context
+
+    Notes
+    -----
+    For a full list of the available protocols and the implementations that
+    they map across to see the latest online documentation:
+
+    - For implementations built into ``fsspec`` see
+      https://filesystem-spec.readthedocs.io/en/latest/api.html#built-in-implementations
+    - For implementations in separate packages see
+      https://filesystem-spec.readthedocs.io/en/latest/api.html#other-known-implementations
     """
     fs, fs_token, paths = get_fs_token_paths(
         urlpath,
@@ -279,9 +289,15 @@ def open_files(
         protocol=protocol,
         expand=expand,
     )
-    if "r" not in mode and auto_mkdir:
+    if fs.protocol == "file":
+        fs.auto_mkdir = auto_mkdir
+    elif "r" not in mode and auto_mkdir:
         parents = {fs._parent(path) for path in paths}
-        [fs.makedirs(parent, exist_ok=True) for parent in parents]
+        for parent in parents:
+            try:
+                fs.makedirs(parent, exist_ok=True)
+            except PermissionError:
+                pass
     return OpenFiles(
         [
             OpenFile(
@@ -301,35 +317,23 @@ def open_files(
 
 
 def _un_chain(path, kwargs):
-    if isinstance(path, (tuple, list)):
-        bits = [_un_chain(p, kwargs) for p in path]
-        out = []
-        for pbit in zip(*bits):
-            paths, protocols, kwargs = zip(*pbit)
-            if len(set(protocols)) > 1:
-                raise ValueError("Protocol mismatch in URL chain")
-            if len(set(paths)) == 1:
-                paths = paths[0]
-            else:
-                paths = list(paths)
-            out.append([paths, protocols[0], kwargs[0]])
-        return out
     x = re.compile(".*[^a-z]+.*")  # test for non protocol-like single word
     bits = (
         [p if "://" in p or x.match(p) else p + "://" for p in path.split("::")]
         if "::" in path
         else [path]
     )
-    if len(bits) < 2:
-        return []
     # [[url, protocol, kwargs], ...]
     out = []
     previous_bit = None
+    kwargs = kwargs.copy()
     for bit in reversed(bits):
-        protocol = split_protocol(bit)[0] or "file"
+        protocol = kwargs.pop("protocol", None) or split_protocol(bit)[0] or "file"
         cls = get_filesystem_class(protocol)
         extra_kwargs = cls._get_kwargs_from_urls(bit)
-        kws = kwargs.get(protocol, {})
+        kws = kwargs.pop(protocol, {})
+        if bit is bits[0]:
+            kws.update(kwargs)
         kw = dict(**extra_kwargs, **kws)
         bit = cls._strip_protocol(bit)
         if (
@@ -363,28 +367,32 @@ def url_to_fs(url, **kwargs):
     urlpath : str
         The file-systems-specific URL for ``url``.
     """
+    # non-FS arguments that appear in fsspec.open()
+    # inspect could keep this in sync with open()'s signature
+    known_kwargs = {
+        "compression",
+        "encoding",
+        "errors",
+        "expand",
+        "mode",
+        "name_function",
+        "newline",
+        "num",
+    }
+    kwargs = {k: v for k, v in kwargs.items() if k not in known_kwargs}
     chain = _un_chain(url, kwargs)
-    if len(chain) > 1:
-        inkwargs = {}
-        # Reverse iterate the chain, creating a nested target_* structure
-        for i, ch in enumerate(reversed(chain)):
-            urls, protocol, kw = ch
-            if i == len(chain) - 1:
-                inkwargs = dict(**kw, **inkwargs)
-                continue
-            inkwargs["target_options"] = dict(**kw, **inkwargs)
-            inkwargs["target_protocol"] = protocol
-            inkwargs["fo"] = urls
-        urlpath, protocol, _ = chain[0]
-        fs = filesystem(protocol, **inkwargs)
-    else:
-        protocol = split_protocol(url)[0]
-        cls = get_filesystem_class(protocol)
-
-        options = cls._get_kwargs_from_urls(url)
-        update_storage_options(options, kwargs)
-        fs = cls(**options)
-        urlpath = fs._strip_protocol(url)
+    inkwargs = {}
+    # Reverse iterate the chain, creating a nested target_* structure
+    for i, ch in enumerate(reversed(chain)):
+        urls, protocol, kw = ch
+        if i == len(chain) - 1:
+            inkwargs = dict(**kw, **inkwargs)
+            continue
+        inkwargs["target_options"] = dict(**kw, **inkwargs)
+        inkwargs["target_protocol"] = protocol
+        inkwargs["fo"] = urls
+    urlpath, protocol, _ = chain[0]
+    fs = filesystem(protocol, **inkwargs)
     return fs, urlpath
 
 
@@ -437,8 +445,18 @@ def open(
     Returns
     -------
     ``OpenFile`` object.
+
+    Notes
+    -----
+    For a full list of the available protocols and the implementations that
+    they map across to see the latest online documentation:
+
+    - For implementations built into ``fsspec`` see
+      https://filesystem-spec.readthedocs.io/en/latest/api.html#built-in-implementations
+    - For implementations in separate packages see
+      https://filesystem-spec.readthedocs.io/en/latest/api.html#other-known-implementations
     """
-    return open_files(
+    out = open_files(
         urlpath=[urlpath],
         mode=mode,
         compression=compression,
@@ -448,10 +466,17 @@ def open(
         newline=newline,
         expand=False,
         **kwargs,
-    )[0]
+    )
+    if not out:
+        raise FileNotFoundError(urlpath)
+    return out[0]
 
 
-def open_local(url, mode="rb", **storage_options):
+def open_local(
+    url: str | list[str] | Path | list[Path],
+    mode: str = "rb",
+    **storage_options: dict,
+) -> str | list[str]:
     """Open file(s) which can be resolved to local
 
     For files which either are local, or get downloaded upon open
@@ -475,7 +500,7 @@ def open_local(url, mode="rb", **storage_options):
         )
     with of as files:
         paths = [f.name for f in files]
-    if isinstance(url, str) and not has_magic(url):
+    if (isinstance(url, str) and not has_magic(url)) or isinstance(url, Path):
         return paths[0]
     return paths
 
@@ -484,7 +509,7 @@ def get_compression(urlpath, compression):
     if compression == "infer":
         compression = infer_compression(urlpath)
     if compression is not None and compression not in compr:
-        raise ValueError("Compression type %s not supported" % compression)
+        raise ValueError(f"Compression type {compression} not supported")
     return compression
 
 
@@ -496,6 +521,8 @@ def split_protocol(urlpath):
         if len(protocol) > 1:
             # excludes Windows paths
             return protocol, path
+    if urlpath.startswith("data:"):
+        return urlpath.split(":", 1)
     return None, urlpath
 
 
@@ -587,62 +614,41 @@ def get_fs_token_paths(
     if isinstance(urlpath, (list, tuple, set)):
         if not urlpath:
             raise ValueError("empty urlpath sequence")
-        urlpath = [stringify_path(u) for u in urlpath]
+        urlpath0 = stringify_path(list(urlpath)[0])
     else:
-        urlpath = stringify_path(urlpath)
-    chain = _un_chain(urlpath, storage_options or {})
-    if len(chain) > 1:
-        inkwargs = {}
-        # Reverse iterate the chain, creating a nested target_* structure
-        for i, ch in enumerate(reversed(chain)):
-            urls, nested_protocol, kw = ch
-            if i == len(chain) - 1:
-                inkwargs = dict(**kw, **inkwargs)
-                continue
-            inkwargs["target_options"] = dict(**kw, **inkwargs)
-            inkwargs["target_protocol"] = nested_protocol
-            inkwargs["fo"] = urls
-        paths, protocol, _ = chain[0]
-        fs = filesystem(protocol, **inkwargs)
-        if isinstance(paths, (list, tuple, set)):
-            paths = [fs._strip_protocol(u) for u in paths]
-        else:
-            paths = fs._strip_protocol(paths)
+        urlpath0 = stringify_path(urlpath)
+    storage_options = storage_options or {}
+    if protocol:
+        storage_options["protocol"] = protocol
+    chain = _un_chain(urlpath0, storage_options or {})
+    inkwargs = {}
+    # Reverse iterate the chain, creating a nested target_* structure
+    for i, ch in enumerate(reversed(chain)):
+        urls, nested_protocol, kw = ch
+        if i == len(chain) - 1:
+            inkwargs = dict(**kw, **inkwargs)
+            continue
+        inkwargs["target_options"] = dict(**kw, **inkwargs)
+        inkwargs["target_protocol"] = nested_protocol
+        inkwargs["fo"] = urls
+    paths, protocol, _ = chain[0]
+    fs = filesystem(protocol, **inkwargs)
+    if isinstance(urlpath, (list, tuple, set)):
+        pchains = [
+            _un_chain(stringify_path(u), storage_options or {})[0] for u in urlpath
+        ]
+        if len({pc[1] for pc in pchains}) > 1:
+            raise ValueError("Protocol mismatch getting fs from %s", urlpath)
+        paths = [pc[0] for pc in pchains]
     else:
-        if isinstance(urlpath, (list, tuple, set)):
-            protocols, paths = zip(*map(split_protocol, urlpath))
-            if protocol is None:
-                protocol = protocols[0]
-                if not all(p == protocol for p in protocols):
-                    raise ValueError(
-                        "When specifying a list of paths, all paths must "
-                        "share the same protocol"
-                    )
-            cls = get_filesystem_class(protocol)
-            optionss = list(map(cls._get_kwargs_from_urls, urlpath))
-            paths = [cls._strip_protocol(u) for u in urlpath]
-            options = optionss[0]
-            if not all(o == options for o in optionss):
-                raise ValueError(
-                    "When specifying a list of paths, all paths must "
-                    "share the same file-system options"
-                )
-            update_storage_options(options, storage_options)
-            fs = cls(**options)
-        else:
-            protocols = split_protocol(urlpath)[0]
-            protocol = protocol or protocols
-            cls = get_filesystem_class(protocol)
-            options = cls._get_kwargs_from_urls(urlpath)
-            paths = cls._strip_protocol(urlpath)
-            update_storage_options(options, storage_options)
-            fs = cls(**options)
-
+        paths = fs._strip_protocol(paths)
     if isinstance(paths, (list, tuple, set)):
         paths = expand_paths_if_needed(paths, mode, num, fs, name_function)
     else:
         if "w" in mode and expand:
             paths = _expand_paths(paths, name_function, num)
+        elif "x" in mode and expand:
+            paths = _expand_paths(paths, name_function, num)
         elif "*" in paths:
             paths = [f for f in sorted(fs.glob(paths)) if not fs.isdir(f)]
         else:
