@@ -5,12 +5,15 @@ import pytest
 
 import fsspec
 from fsspec.implementations.local import LocalFileSystem
-from fsspec.implementations.reference import ReferenceFileSystem, ReferenceNotReachable
+from fsspec.implementations.reference import (
+    LazyReferenceMapper,
+    ReferenceFileSystem,
+    ReferenceNotReachable,
+)
 from fsspec.tests.conftest import data, realfile, reset_files, server, win  # noqa: F401
 
 
 def test_simple(server):  # noqa: F811
-
     refs = {
         "a": b"data",
         "b": (realfile, 0, 5),
@@ -53,6 +56,16 @@ def test_ls(server):  # noqa: F811
     }
 
 
+def test_nested_dirs_ls():
+    # issue #1430
+    refs = {"a": "A", "B/C/b": "B", "B/C/d": "d", "B/_": "_"}
+    fs = fsspec.filesystem("reference", fo=refs)
+    assert len(fs.ls("")) == 2
+    assert {e["name"] for e in fs.ls("")} == {"a", "B"}
+    assert len(fs.ls("B")) == 2
+    assert {e["name"] for e in fs.ls("B")} == {"B/C", "B/_"}
+
+
 def test_info(server):  # noqa: F811
     refs = {
         "a": b"data",
@@ -338,8 +351,8 @@ def test_multi_fs_provided(m, tmpdir):
     # local URLs are file:// by default
     refs = {
         "a": b"data",
-        "b": ("file://" + str(real), 0, 5),
-        "c/d": ("file://" + str(real), 1, 6),
+        "b": (f"file://{real}", 0, 5),
+        "c/d": (f"file://{real}", 1, 6),
         "c/e": ["memory://afile"],
     }
 
@@ -361,8 +374,8 @@ def test_multi_fs_created(m, tmpdir):
     # local URLs are file:// by default
     refs = {
         "a": b"data",
-        "b": ("file://" + str(real), 0, 5),
-        "c/d": ("file://" + str(real), 1, 6),
+        "b": (f"file://{real}", 0, 5),
+        "c/d": (f"file://{real}", 1, 6),
         "c/e": ["memory://afile"],
     }
 
@@ -408,7 +421,7 @@ def test_fss_has_defaults(m):
 
     fs = fsspec.filesystem("reference", fs={"memory": m}, fo={})
     assert fs.fss["memory"] is m
-    assert fs.fss[None].protocol == "file"
+    assert fs.fss[None].protocol == ("file", "local")
 
     fs = fsspec.filesystem("reference", fs={None: m}, fo={})
     assert fs.fss[None] is m
@@ -440,6 +453,7 @@ def test_merging(m):
 def test_cat_file_ranges(m):
     other = b"other test data"
     m.pipe("/b", other)
+
     fs = fsspec.filesystem(
         "reference",
         fo={
@@ -458,15 +472,26 @@ def test_cat_file_ranges(m):
     assert fs.cat_file("d", 1, -3) == other[4:10][1:-3]
 
 
-def test_cat_missing(m):
+@pytest.mark.parametrize(
+    "fo",
+    [
+        {
+            "c": ["memory://b"],
+            "d": ["memory://unknown", 4, 6],
+        },
+        {
+            "c": ["memory://b"],
+            "d": ["//unknown", 4, 6],
+        },
+    ],
+    ids=["memory protocol", "mixed protocols: memory and unspecified"],
+)
+def test_cat_missing(m, fo):
     other = b"other test data"
     m.pipe("/b", other)
     fs = fsspec.filesystem(
         "reference",
-        fo={
-            "c": ["memory://b"],
-            "d": ["memory://unknown", 4, 6],
-        },
+        fo=fo,
     )
     with pytest.raises(FileNotFoundError):
         fs.cat("notafile")
@@ -499,6 +524,11 @@ def test_cat_missing(m):
     out = mapper.getitems(["c", "d"], on_error="return")
     assert isinstance(out["d"], ReferenceNotReachable)
 
+    out = fs.cat(["notone", "c", "d"], on_error="return")
+    assert isinstance(out["notone"], FileNotFoundError)
+    assert out["c"] == other
+    assert isinstance(out["d"], ReferenceNotReachable)
+
     out = mapper.getitems(["c", "d"], on_error="omit")
     assert list(out) == ["c"]
 
@@ -621,3 +651,69 @@ def test_mapping_getitems(m):
     fs = fsspec.filesystem("reference", fo=refs, fs=h)
     mapping = fs.get_mapper("")
     assert mapping.getitems(["b", "a"]) == {"a": b"A", "b": b"B"}
+
+
+def test_cached(m, tmpdir):
+    fn = f"{tmpdir}/ref.json"
+
+    m.pipe({"a": b"A", "b": b"B"})
+    m.pipe("ref.json", b"""{"a": ["a"], "b": ["b"]}""")
+
+    fs = fsspec.filesystem(
+        "reference",
+        fo="simplecache::memory://ref.json",
+        fs=m,
+        target_options={"cache_storage": str(tmpdir), "same_names": True},
+    )
+    assert fs.cat("a") == b"A"
+    assert os.path.exists(fn)
+
+    # truncate original file to show we are loading from the cached version
+    m.pipe("ref.json", b"")
+    fs = fsspec.filesystem(
+        "reference",
+        fo="simplecache::memory://ref.json",
+        fs=m,
+        target_options={"cache_storage": str(tmpdir), "same_names": True},
+        skip_instance_cache=True,
+    )
+    assert fs.cat("a") == b"A"
+
+
+@pytest.fixture()
+def lazy_refs(m):
+    zarr = pytest.importorskip("zarr")
+    l = LazyReferenceMapper.create("memory://refs", fs=m)
+    g = zarr.open(l, mode="w")
+    g.create_dataset(name="data", shape=(100,), chunks=(10,), dtype="int64")
+    return l
+
+
+def test_append_parquet(lazy_refs, m):
+    with pytest.raises(KeyError):
+        lazy_refs["data/0"]
+    lazy_refs["data/0"] = b"data"
+    assert lazy_refs["data/0"] == b"data"
+    lazy_refs.flush()
+
+    lazy2 = LazyReferenceMapper("memory://refs", fs=m)
+    assert lazy2["data/0"] == b"data"
+    with pytest.raises(KeyError):
+        lazy_refs["data/1"]
+    lazy2["data/1"] = b"Bdata"
+    assert lazy2["data/1"] == b"Bdata"
+    lazy2.flush()
+
+    lazy2 = LazyReferenceMapper("memory://refs", fs=m)
+    assert lazy2["data/0"] == b"data"
+    assert lazy2["data/1"] == b"Bdata"
+    lazy2["data/1"] = b"Adata"
+    del lazy2["data/0"]
+    assert lazy2["data/1"] == b"Adata"
+    assert "data/0" not in lazy2
+    lazy2.flush()
+
+    lazy2 = LazyReferenceMapper("memory://refs", fs=m)
+    with pytest.raises(KeyError):
+        lazy2["data/0"]
+    assert lazy2["data/1"] == b"Adata"
