@@ -1,4 +1,5 @@
 import contextlib
+import gzip
 import json
 import os
 import threading
@@ -10,7 +11,7 @@ import pytest
 requests = pytest.importorskip("requests")
 port = 9898
 data = b"\n".join([b"some test data"] * 1000)
-realfile = "http://127.0.0.1:%i/index/realfile" % port
+realfile = f"http://127.0.0.1:{port}/index/realfile"
 index = b'<a href="%s">Link</a>' % realfile.encode()
 listing = open(
     os.path.join(os.path.dirname(__file__), "data", "listing.html"), "rb"
@@ -56,11 +57,14 @@ class HTTPTestHandler(BaseHTTPRequestHandler):
         file_data = self.files.get(file_path)
         if "give_path" in self.headers:
             return self._respond(200, data=json.dumps({"path": self.path}).encode())
+        if "redirect" in self.headers and file_path != "/index/realfile":
+            new_url = f"http://127.0.0.1:{port}/index/realfile"
+            return self._respond(301, {"Location": new_url})
         if file_data is None:
             return self._respond(404)
 
         status = 200
-        content_range = "bytes 0-%i/%i" % (len(file_data) - 1, len(file_data))
+        content_range = f"bytes 0-{len(file_data) - 1}/{len(file_data)}"
         if ("Range" in self.headers) and ("ignore_range" not in self.headers):
             ran = self.headers["Range"]
             b, ran = ran.split("=")
@@ -71,15 +75,26 @@ class HTTPTestHandler(BaseHTTPRequestHandler):
             else:
                 # suffix only
                 l = len(file_data)
-                content_range = f"bytes {l-int(end)}-{l-1}/{l}"
+                content_range = f"bytes {l - int(end)}-{l - 1}/{l}"
                 file_data = file_data[-int(end) :]
             if "use_206" in self.headers:
                 status = 206
         if "give_length" in self.headers:
-            response_headers = {"Content-Length": len(file_data)}
+            if "gzip_encoding" in self.headers:
+                file_data = gzip.compress(file_data)
+                response_headers = {
+                    "Content-Length": len(file_data),
+                    "Content-Encoding": "gzip",
+                }
+            else:
+                response_headers = {"Content-Length": len(file_data)}
             self._respond(status, response_headers, file_data)
         elif "give_range" in self.headers:
             self._respond(status, {"Content-Range": content_range}, file_data)
+        elif "give_mimetype" in self.headers:
+            self._respond(
+                status, {"Content-Type": "text/html; charset=utf-8"}, file_data
+            )
         else:
             self._respond(status, data=file_data)
 
@@ -123,11 +138,15 @@ class HTTPTestHandler(BaseHTTPRequestHandler):
             response_headers = {"Content-Length": len(file_data)}
             if "zero_length" in self.headers:
                 response_headers["Content-Length"] = 0
+            elif "gzip_encoding" in self.headers:
+                file_data = gzip.compress(file_data)
+                response_headers["Content-Encoding"] = "gzip"
+                response_headers["Content-Length"] = len(file_data)
 
             self._respond(200, response_headers)
         elif "give_range" in self.headers:
             self._respond(
-                200, {"Content-Range": "0-%i/%i" % (len(file_data) - 1, len(file_data))}
+                200, {"Content-Range": f"0-{len(file_data) - 1}/{len(file_data)}"}
             )
         elif "give_etag" in self.headers:
             self._respond(200, {"ETag": "xxx"})
@@ -143,7 +162,7 @@ def serve():
     th.daemon = True
     th.start()
     try:
-        yield "http://127.0.0.1:%i" % port
+        yield f"http://127.0.0.1:{port}"
     finally:
         httpd.socket.close()
         httpd.shutdown()
